# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

__all__ = ['find_zetup_config', 'Zetup',
  'ZetupError', 'ZetupConfigNotFound',
  'DistributionNotFound', 'VersionConflict',
  'Packages', 'Package',
  'Popen', 'call']

import sys
import os
from types import ModuleType
## from inspect import ismodule

from .zetup import Zetup
from .error import ZetupError
from .config import ZetupConfigNotFound
from .requires import DistributionNotFound, VersionConflict
from .package import Packages, Package
from .process import Popen, call


COMMANDS = {}


def command(func):
    """Decorator for registering basic (non-project-bound) commands.
    """
    COMMANDS[func.__name__] = func


def find_zetup_config(pkgname):
    zfg_modname = pkgname + '.zetup_config'
    try: # Already imported?
        return sys.modules[zfg_modname]
    except KeyError:
        pass
    try:
        return __import__(zfg_modname).zetup_config
    except ImportError:
        pass
    # ==> no zetup config module
    # ==> assume package imported from source (repo)
    # ==> load setup config from package's parent path:
    mod = sys.modules[pkgname]
    path = os.path.dirname(os.path.dirname(os.path.realpath(mod.__file__)))
    try:
        return Zetup(path)
    except ZetupConfigNotFound as e:
        raise ZetupConfigNotFound(
          "No '%s.zetup_config' module and: %s" % (pkgname, e))


def annotate(pkgname, check_requirements=True, check_packages=True):
    """Find the zetup config for given `pkgname`
       and add __version__, __requires__, __dist__, __description__,
       __packages__ and __extras__ (if defined) to the package object.

    - Automatically checks installed package requirements
      unless `check_requirements` is False.
    - Automatically checks installed package files
      unless `check_packages` is False.
    - Returns the zetup config object.
    """
    try:
        mod = sys.modules[pkgname]
    except KeyError:
        raise ZetupError(
          "Package %s was not found in sys.modules" % repr(pkgname))
    zfg = find_zetup_config(pkgname)
    mod.__version__ = zfg.VERSION
    mod.__requires__ = zfg.REQUIRES
    if check_requirements:
        zfg.REQUIRES.check()
    if zfg.EXTRAS:
        mod.__extras__ = zfg.EXTRAS
    mod.__distribution__ = zfg.DISTRIBUTION.find(os.path.dirname(__file__))
    mod.__description__ = zfg.DESCRIPTION
    mod.__packages__ = zfg.PACKAGES
    if (check_packages
        #TODO: remove (only for backwards compatibility)
        and isinstance(zfg.PACKAGES, Packages)
        ):
        zfg.PACKAGES.check()
    return zfg


annotate(__name__, check_requirements=False)


class ZetupModule(ModuleType):
    """Custom module class for wrapping native ``zetup`` module on import,
       to dynamically provide additional module members
       based on installed dependencies.
    """
    def __init__(self):
        self.__name__ = __name__
        self.module = sys.modules[__name__]
        try:
            from .path import Path
        except ImportError: #==> no path.py installed
            pass
        else:
            self.__all__.append('Path')

    def __getattr__(self, name):
        """Just delegate to native ``zetup`` module.
        """
        return getattr(self.module, name)

    def __dir__(self):
        """Just delegate to native ``zetup`` module.
        """
        return list(self.__dict__) + dir(self.module)

    @property
    def Path(self):
        """Provide **path.py**-wrapping ``zetup.Path`` class
           if **path.py** is installed.
        """
        from .path import Path
        return Path

# replace native zetup module with wrapper
sys.modules[__name__] = ZetupModule()


# # Get zetup's own config:
# zfg = find_zetup_config(__name__)

# __distribution__ = zfg.DISTRIBUTION.find(__path__[0])
# __description__ = zfg.DESCRIPTION

# __version__ = zfg.VERSION

# __requires__ = zfg.REQUIRES # .checked
# ## if ismodule(zfg): # if this is an installed zetup package:
# ##     __requires__.check()

# __extras__ = zfg.EXTRAS

# ## __notebook__ = zfg.NOTEBOOKS['README']


def setup_entry_point(dist, keyword, value):
    """zetup's ``entry_point`` handler for ``setup()``
       in a project's **setup.py**, setting all setup keyword parameters
       based on zetup config.

    - Activated with ``setup(setup_requires=['zetup'], use_zetup=True)``
    """
    assert keyword == 'use_zetup'
    if not value:
        return

    zetup = Zetup() # reads config from current working directory
                    # (where setup.py is run)
    keywords = zetup.setup_keywords()
    for name, value in keywords.items():
        # Generally, setting stuff on dist.metadata is enough
        # (and necessary), but *pip* only works correct
        # if stuff is also set directly on dist object
        # (seems to read at least packages list somehow from there):
        for obj in [dist, dist.metadata]:
            setattr(obj, name, value)

    # finally run any custom setup hooks defined in project's zetup config,
    # but first check setup requirements
    if zetup.SETUP_REQUIRES:
        zetup.SETUP_REQUIRES.check()
    if zetup.SETUP_HOOKS:
        sys.path.insert(0, '.')
        for hook in zetup.SETUP_HOOKS:
            modname, funcname = hook.split(':')
            mod = __import__(modname)
            for subname in modname.split('.')[1:]:
                mod = getattr(mod, subname)
            func = getattr(mod, funcname)
            func(dist)



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

__all__ = ['Path']

from path import Path


class Path(Path):
    """A little wrapper for **path.py** providing a :func:`.samefile` wrapper
       with Windows Python 2 compatibility.
    """
    def samefile(self, other):
        """Also working if there is no ``os.path.samefile()``.
        """
        if not hasattr(self.module, 'samefile'):
            other = Path(other).realpath().normpath().normcase()
            return self.realpath().normpath().normcase() == other
        return self.module.samefile(self, other)



"""Test the zetup package.

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
import sys
from inspect import ismodule

from pkg_resources import (
  parse_requirements, get_distribution, DistributionNotFound)

import zetup
from zetup import Zetup, Path

import pytest


def test_zetup_module(zfg, in_repo, in_site_packages):
    """Test the zetup module wrapper,
       which dynamically provides additional members
       based on installed dependencies.
    """
    assert ismodule(zetup)
    assert zetup.__class__.__module__ == 'zetup'
    assert set(zetup.__all__) == set(zetup.module.__all__).union(['Path'])


def test_zetup_config(zfg, in_repo, in_site_packages):
    """Test the zetup config management on zetup package's own config.
    """
    #--- NAME ---
    assert zfg.NAME == 'zetup'

    pkg_path = Path(zetup.__path__[0]).realpath()

    if in_site_packages:
        zfg_path = Path(zfg.__file__).realpath().dirname()
    else:
        zfg_path = Path(zfg.ZETUP_DIR).realpath()

    if in_site_packages:
        assert zfg.__name__ == 'zetup.zetup_config'
        assert zfg_path == pkg_path # / 'zetup_config'
    else:
        with pytest.raises(ImportError):
            from zetup import zetup_config

        assert isinstance(zfg, Zetup)
        assert zfg_path == pkg_path.dirname()

    #--- DESCRIPTION ---
    assert zetup.__description__ == zfg.DESCRIPTION

    #--- DISTRIBUTION ---
    try:
        dist = get_distribution('zetup')
    except DistributionNotFound:
        if in_repo:
            dist = None
        else:
            raise
    else:
        if in_repo and not Path(dist.location).samefile(pkg_path.dirname()):
            dist = None
    assert zetup.__distribution__ == zfg.DISTRIBUTION.find(pkg_path) \
      == dist

    #--- VERSION ---
    assert zetup.__version__ == zfg.VERSION

    #--- REQUIRES ---
    assert zetup.__requires__ == zfg.REQUIRES
    if in_repo:
        assert list(parse_requirements(str(zetup.__requires__))) == list(
          parse_requirements((zfg_path / 'requirements.txt').text()))

    #--- EXTRAS ---
    for extra in zetup.__extras__:
        assert zetup.__extras__[extra] == zfg.EXTRAS[extra]
        if in_repo:
            assert list(parse_requirements(str(zetup.__extras__[extra]))) \
              == list(parse_requirements(
                   (zfg_path / ('requirements.%s.txt' % extra))
                   .text()))

    #--- PYTHON ---
    assert ('%s.%s' % sys.version_info[:2]) in zfg.PYTHON

    #--- CLASSIFIERS ---
    assert all('Programming Language :: Python :: %s' % version
               for version in zfg.PYTHON)



"""Test zetup's **path.py** wrapper.

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
import sys
import os

WIN = sys.platform.startswith('win')

from path import Path

import zetup


def test_path():
    """Test zetup's **path.py** wrapper using ``__file__``.
    """
    assert Path in zetup.Path.mro()
    assert zetup.Path.module is os.path

    # check that Path.samefile() is working if no os.path.samefile exists
    # (like in Windows Python 2.x)
    try:
        samefile_func = os.path.samefile
        delattr(os.path, 'samefile')
    except AttributeError:
        samefile_func = None
    path = zetup.Path(__file__)
    relpath = os.path.relpath(__file__, os.getcwd())
    if WIN: # also check case insensitivity
        path = zetup.Path(path.upper())
        relpath = relpath.lower()
    for relpath in [relpath, Path(relpath), zetup.Path(relpath)]:
        assert path.samefile(relpath)
    if samefile_func:
        os.path.samefile = samefile_func



"""Test the zetup.notebook.Notebook class.

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
from inspect import getmembers

from IPython import nbconvert

import pytest


def test_zetup_notebook(zfg, in_site_packages):
    """Test zetup.Notebook features using zetup's own README notebook.
    """
    if in_site_packages:
        # notebooks aren't installed
        return

    notebook = zfg.NOTEBOOKS['README']

    #--- CONVERTERS ---

    # check if all IPython.nbconvert.export_...() converters
    # are exposed to the notebook instance as to_...() methods
    for name, obj in getmembers(nbconvert):
        if name.startswith('export_'):
            converter = getattr(notebook, 'to_' + name.split('_', 1)[1])
            assert callable(converter)

    # check correct AttributeError messages
    # (if a non-existing attribute starts with 'to_',
    #  __getattr__ should report that such a 'converter'
    #  instead of 'attribute' does not exist)
    with pytest.raises(AttributeError) as exc:
        notebook.not_defined
    assert all(text in str(exc.value) for text in [
      "has no attribute", "not_defined"])
    assert "converter" not in str(exc.value)
    with pytest.raises(AttributeError) as exc:
        notebook.to_not_defined()
    assert all(text in str(exc.value) for text in [
      "has no converter", "to_not_defined"])



"""Common fixtures for zetup tests.

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
from inspect import ismodule

from zetup import find_zetup_config

import pytest


@pytest.fixture
def zfg():
    """Get zetup's own config.
    """
    return find_zetup_config('zetup')


@pytest.fixture
def in_site_packages():
    """Is zetup imported from repository?
    """
    return ismodule(zfg())


@pytest.fixture
def in_repo():
    """Is zetup imported from site-packages?
    """
    return not in_site_packages()



from __future__ import print_function

import sys
import os

## from setuptools import Distribution
from pkg_resources import get_distribution, working_set, VersionConflict


def samefile(path, other):
    return os.path.normcase(os.path.normpath(os.path.realpath(path))) \
      == os.path.normcase(os.path.normpath(os.path.realpath(other)))


try:
    import zetup
except VersionConflict:
    egg_info = 'zetup.egg-info'
    dist = get_distribution('zetup')
    if samefile(dist.location, os.path.dirname(os.path.realpath(__file__))) \
      and os.path.exists(egg_info):
        print("zetup: Removing possibly outdated %s/" % egg_info)
        for fname in os.listdir(egg_info):
            os.remove(os.path.join(egg_info, fname))
        os.rmdir(egg_info)
        # when run via pip, the egg-info is still referenced by setuptools,
        # which would try to read the contents
        for keys in working_set.entry_keys.values():
            if 'zetup' in keys:
                keys.remove('zetup')
        del working_set.by_key['zetup']


sys.path.insert(0, os.path.realpath(os.path.dirname(__file__)))
from zetup import Zetup, DistributionNotFound, VersionConflict
try:
    import zetup.commands
except (ImportError, DistributionNotFound, VersionConflict):
    #==> no zetup commands available
    # standard setup commands work anyway
    pass


## setup_req = 'setuptools >= 15.0'
## try:
##     get_distribution(setup_req)
## except VersionConflict:
##     for mod in ['setuptools', 'pkg_resources']:
##         for name, _ in list(sys.modules.items()):
##             if name == mod or name.startswith(mod + '.'):
##                 del sys.modules[name]
##     sys.path.insert(0, Distribution().fetch_build_egg(setup_req))


zfg = Zetup()

setup = zfg.setup
setup['package_data']['zetup.commands.make'] = [
  'templates/*.jinja',
  'templates/package/*.jinja',
  ]
setup(
  ## setup_requires=setup_req,
  ## setup_requires=['hgdistver >= 0.25'],

  ## get_version_from_scm=True,

  entry_points={
    'distutils.setup_keywords': [
      'use_zetup = zetup:setup_entry_point',
      ],
    'console_scripts': [
      'zetup = zetup.script:run',
      ]},
  )



#!python

# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function

import sys
from itertools import chain
from argparse import ArgumentParser
import distutils.command

import zetup.commands
from zetup.commands import ZetupCommandError
from zetup.process import call


EXTERNAL_COMMANDS = []

COMMANDS = sorted(chain(
  distutils.command.__all__,
  zetup.COMMANDS,
  zetup.Zetup.COMMANDS,
  EXTERNAL_COMMANDS,
  ))

ap = ArgumentParser()
ap.add_argument(
  'cmd', choices=COMMANDS,
  help="command",
  )

args = ap.parse_args()


def run():
    exit_status = 0 # exit status of this script
    try:
        cmdfunc = zetup.COMMANDS[args.cmd]
    except KeyError:
        if args.cmd in EXTERNAL_COMMANDS:
            exit_status = call([args.cmd])
        else:
            zfg = zetup.Zetup()
            if args.cmd in zfg.COMMANDS:
                try:
                    exit_status = getattr(zfg, args.cmd)()
                except ZetupCommandError as e:
                    print("Error: %s" % e, file=sys.stderr)
                    exit_status = 1
                else:
                    try: # return value can be more than just a status number
                        exit_status = exit_status.status
                    except AttributeError:
                        pass
            else: # ==> standard setup command
                zfg(subprocess=True)
    else:
        exit_status = cmdfunc()

    sys.exit(exit_status or 0)


if __name__ == '__main__':
    run()
